// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <i.h>
#include <math.h>
#include "../../sim.h"
#include "../../dump.h"
#include "../../elec_plugins.h"
#include "../../ntricks.h"
#include "stark.h"
#include "../../dynamic_store.h"

static int unused __attribute__((unused));

struct stark stark_config;

void stark_cal_stark(double *avg_mode,struct device *in,struct istruct *mode,double *phi_in)
{
	struct istruct E;
	struct istruct phi;
	inter_import_array(&E,in->ymesh, phi_in,in->ymeshpoints,TRUE);
	inter_import_array(&phi,in->ymesh, phi_in,in->ymeshpoints,TRUE);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"stark_phi.dat");
	inter_deriv(&E,&phi);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"stark_E.dat");
	inter_pow(&E,2.0);
	inter_mul(&E,stark_config.stark_stark);
	if (get_dump_status(dump_iodump)==TRUE) inter_save_a(&E,in->outputpath,"stark_x_stark.dat");

	inter_convolve(&E,mode);
	*avg_mode=inter_avg(&E);

	inter_free(&E);
	inter_free(&phi);
}

void get_avg_np_pos(struct device *in,double *nx,double *px)
{
int i;
double navg=0.0;
double pavg=0.0;
double nsum=0.0;
double psum=0.0;


for (i=0;i<in->ymeshpoints;i++)
{
	navg+=(in->n[i]+in->nt_all[i]-in->n_orig[i])*in->ymesh[i];
	pavg+=(in->p[i]+in->pt_all[i]-in->p_orig[i])*in->ymesh[i];
	nsum+=(in->n[i]+in->nt_all[i]-in->n_orig[i]);
	psum+=in->p[i]+in->pt_all[i]-in->p_orig[i];

}
if (nsum!=0.0)
{
*nx=navg/nsum;
//printf("%le\n",navg/nsum);
}else
{
*nx=0.0;
}

if (psum!=0.0)
{
*px=pavg/psum;
}else
{
*px=0.0;
}



}

double stark_get_light_den(struct device *in)
{
int i=0;
double tot=0.0;
for (i=0;i<in->ymeshpoints;i++)
{
	tot+=in->Gn[i];
}
tot/=(double)in->ymeshpoints;

return tot;
}

void stark_update_alpha(struct device *in,struct light *mylight,struct istruct *alpha_reduce,struct istruct *n_reduce,int lam)
{
int i;
//int lam;
double pos=0.0;
double dn=0.0;
double reduce_alpha=0.0;
double reduce_n=0.0;
dn=0.0;

for (i=0;i<in->ymeshpoints;i++)
{
	dn+=(in->n[i]+in->p[i])/(in->nf_save[i]+in->pf_save[i]);
	//printf("%lf\n",dn);
}

dn/=((double)in->ymeshpoints);

for (i=0;i<in->ymeshpoints;i++)
{


		reduce_alpha=(dn-1.0)*stark_config.stark_saturate;
		reduce_alpha=(1.0-reduce_alpha);
		if (reduce_alpha<1e-4) reduce_alpha=1e-4;

		alpha_reduce->data[i]=reduce_alpha;

		reduce_n=(dn-1.0)*stark_config.stark_saturate_n;
		reduce_n=(1.0-reduce_n);
		if (reduce_n<0.33) reduce_n=0.33;
		if (reduce_n>4.0) reduce_n=4.0;
		n_reduce->data[i]=reduce_n;
}

	//for (lam=0;lam<mylight->lpoints;lam++)
	//{
		for (i=0;i<mylight->points;i++)
		{
			if ((mylight->x[i]>mylight->device_start)&&(mylight->x[i]<(mylight->device_start+mylight->device_ylen)))
			{
				//if (lam==0) printf("%le\n",mylight->n[lam][i]);


				pos=mylight->x[i]-mylight->device_start;
				//printf("here %lf\n",pos);
				reduce_alpha=inter_get_noend(alpha_reduce,pos);

				reduce_n=inter_get_noend(n_reduce,pos);
				//printf("%le %le %le\n",pos,reduce_alpha,reduce_n);
				mylight->alpha[lam][i]=mylight->alpha0[lam][i]*reduce_alpha;
				mylight->n[lam][i]=mylight->n[lam][i]*reduce_n;

			}
		}
	//}
return;
}

double stark_get_equiv_laser_power(struct device *in)
{
double laserpower=0.0;
double e0=0.0;
double delta=5.0;
double e1=0.0;
double deriv=0.0;
light_solve_and_update(in,&(in->mylight),in->Psun,laserpower,stark_config.stark_laser_width);
e0=stark_config.stark_den-stark_get_light_den(in);
laserpower+=delta;
light_solve_and_update(in,&(in->mylight),in->Psun,laserpower,stark_config.stark_laser_width);
e1=stark_config.stark_den-stark_get_light_den(in);
deriv=(e1-e0)/delta;
delta=-e1/deriv;
printf("delta=%le laserpower=%le\n",delta,laserpower);
laserpower+=delta;

do
{
	e0=e1;
	light_solve_and_update(in,&(in->mylight),in->Psun,laserpower,stark_config.stark_laser_width);
	e1=stark_config.stark_den-stark_get_light_den(in);
	deriv=(e1-e0)/delta;
	delta=-e1/deriv;
	printf("delta=%le laserpower=%le\n",delta,laserpower);
	laserpower+=delta;

}while(e1>1e-10);
printf("set density=%le wanted density=%le\n",stark_get_light_den(in),stark_config.stark_den);
light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
return laserpower;
}

void sim_stark(struct device *in)
{
struct buffer buf;
buffer_init(&buf);

stark_load_config(&stark_config,in);

struct istruct mymode;
struct light modes;

struct istruct alpha_reduce;
struct istruct n_reduce;


inter_import_array(&alpha_reduce,in->ymesh,in->ymesh,in->ymeshpoints,TRUE);
inter_import_array(&n_reduce,in->ymesh,in->ymesh,in->ymeshpoints,TRUE);

double stark_mode=0.0;
int i;
double nx=0.0;
double px=0.0;

double mode_max=0.0;

//getchar();

in->dt=stark_config.stark_dt_neg_time;

carrier_count_reset(in);

in->go_time=FALSE;

printf("Running stark. %e\n",stark_config.stark_externalv);
int ittr=0;
int pittr=-1;
int step=0;

light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);

//printf("%lf\n",in->Vapplied);
//newton_sim_stark(in,in->Vapplied,FALSE);



newton_sim_stark_simple(in);

if (fabs(stark_config.stark_externalv-in->Vapplied)>0.2)
{
	ramp(in,0.0,stark_config.stark_externalv,7.0);
}

sim_externalv(in,stark_config.stark_externalv);

remesh_reset(in,stark_config.stark_externalv);


newton_sim_stark(in,stark_config.stark_externalv,FALSE);
//newton_sim_stark(in);

time_init(in);

light_init(&modes,in,in->outputpath);
inter_import_array(&mymode,in->ymesh,in->ymesh,in->ymeshpoints,TRUE);
light_set_dx(&modes,in->ymesh[1]-in->ymesh[0]);
light_load_config(&modes);

int probe_mode_pos=light_find_wavelength(&modes,stark_config.stark_probe_wavelength);

light_set_unity_laser_power(&modes,probe_mode_pos);

light_solve_all(&modes);

light_cal_photon_density(&modes);

light_dump(&modes);

light_get_mode(&mymode,probe_mode_pos,&modes);

mode_max=inter_get_max(&mymode);

inter_div_double(&mymode,mode_max);
inter_save_a(&mymode,in->outputpath,"stark_mode_probe.dat");


device_timestep(in);

in->go_time=TRUE;

in->time=stark_config.stark_tstart;

	FILE *out=fopen("./stark.dat","w");
	fclose(out);

struct istruct starkoutn;
inter_init(&starkoutn);

struct istruct starkoutp;
inter_init(&starkoutp);

struct istruct starkoutj;
inter_init(&starkoutj);

struct istruct starkoutstark_mode;
inter_init(&starkoutstark_mode);

struct dynamic_store store;
dump_dynamic_init(&store,in);

double fifteen_volt_field=0.0;
double no_field=0.0;
double stark_mul=0.0;
double steady_exp_value=0.0;
stark_cal_stark(&no_field,in,&mymode,in->phi);


double *high_voltage_phi;

high_voltage_phi=(double*)malloc(in->ymeshpoints*sizeof(double));

double pos=(in->phi[0]-15.0);
double delta=(in->phi[in->ymeshpoints-1]-pos)/((double)(in->ymeshpoints-1));
for (i=0;i<in->ymeshpoints;i++)
{
	high_voltage_phi[i]=pos;
	pos+=delta;
	//printf("%le %le %le\n",in->ymesh[i],in->phi[i],high_voltage_phi[i]);
}

stark_cal_stark(&fifteen_volt_field,in,&mymode,high_voltage_phi);

free(high_voltage_phi);


struct istruct steady_stark;
inter_load(&steady_stark,"steady_stark.inp");
inter_sort(&steady_stark);
steady_exp_value=inter_get_noend(&steady_stark,stark_config.stark_probe_wavelength);

inter_free(&steady_stark);



//stark_mode=steady_exp_value*(stark_mode-no_field)/(fifteen_volt_field-no_field);

//pos=(in->phi[0]-15.0);
//delta=(in->phi[in->ymeshpoints-1]-pos)/((double)(in->ymeshpoints-1));
//for (i=0;i<in->ymeshpoints;i++)
//{
	//in->phi[i]=pos;
	//pos+=delta;
	//printf("%le %le %le\n",in->ymesh[i],in->phi[i],high_voltage_phi[i]);
//}

//stark_cal_stark(&stark_mode,in,&mymode,in->phi);
//stark_mode=steady_exp_value*(stark_mode-no_field)/(fifteen_volt_field-no_field);
//printf("%le\n",stark_mode);
//printf("%le %le %le %le\n",field,no_field,field-no_field);


//getchar();


printf("Main stark loop\n");
double stark=0.0;

double cur;
reset_np_save(in);
set_orig_charge_den(in);
double n_start=(get_free_n_charge(in)+get_free_p_charge(in))/2;
reset_np_save(in);
do
{
in->ilast=get_I(in)+in->C*(in->Vapplied-in->Vapplied_last)/in->dt+in->Vapplied/in->Rshunt;

cur=newton_sim_stark(in,stark_config.stark_externalv,TRUE);
//newton_sim_stark_simple(in);

//(in->time+in->dt>0.0)&&(in->time<=0.0)
if (in->time>=-in->dt/2.0)
{
	pittr++;
}


if ((pittr>=0)&&(pittr<=1))
{
		in->dt=stark_config.stark_dt;


			light_solve_and_update(in,&(in->mylight),in->Psun,stark_config.stark_lasereff,stark_config.stark_laser_width);
			//if (in->mylight.function==0)
			{
				for (i=0;i<in->ymeshpoints;i++)
				{
					in->Gn[i]*=stark_config.stark_laser_width/in->dt;
					in->Gp[i]*=stark_config.stark_laser_width/in->dt;
					//printf("%le\n",in->mylight.pulseL/in->dt);
				}

			}

/*	if (get_dump_status(dump_iodump)==TRUE)
	{
		sprintf(outpath,"%s%s",in->outputpath,"./stark_exciteprofile.dat");
		out=fopen(outpath,"w");
		for (i=0;i<in->ymeshpoints;i++)
		{
			fprintf(out,"%le %le\n",in->ymesh[i],in->Gn[i]);
		}
		fclose(out);
	}*/
}

if (pittr>1)
{

	for (i=0;i<in->ymeshpoints;i++)
	{
		in->Gn[i]=0.0;
		in->Gp[i]=0.0;
	}
	in->avg_gen=get_charge_delta(in);

}

	dump_dynamic_add_data(&store,in,in->time);

	dump_write_to_disk(in);

	if ((stark_config.stark_saturate_n!=0.0)||(stark_config.stark_saturate!=0.0))
	{

		//printf("update alpha and n\n");
		//stark_update_alpha(in,&modes,&alpha_reduce,&n_reduce,probe_mode_pos);

		//printf("%le %le\n",reduce_alpha,reduce_n);
		//light_calculate_complex_n(&modes);

		//light_solve_all(&modes);
		//light_solve_lam_slice(&modes,probe_mode_pos);
		//light_cal_photon_density(&modes);

		//light_get_mode(&mymode,probe_mode_pos,&modes);

		//inter_div_double(&mymode,mode_max);

		//inter_save_a(&alpha_reduce,in->outputpath,"alpha_reduce.dat");
		//inter_save_a(&n_reduce,in->outputpath,"n_reduce.dat");
		//getchar();
		//inter_save_a(&mymode,in->outputpath,"stark_mode_probe.dat");
		//getchar();
	}
	stark_cal_stark(&stark_mode,in,&mymode,in->phi);
	stark_mode=steady_exp_value*(stark_mode-no_field)/(fifteen_volt_field-no_field);
	//stark_mode=stark_mode*;
	//printf("%le %le\n",stark_mode,(1.0-reduce));
	//stark=stark_stark(in);



	ittr++;

	plot_now(in,"stark.plot");
	if (pittr>=1)
	{
	get_avg_np_pos(in,&nx,&px);

	//printf("%le %le\n",nx,px);
	inter_append(&starkoutn,in->time,nx);
	inter_append(&starkoutp,in->time,px);
	}

	inter_append(&starkoutstark_mode,in->time,stark_mode);
	//getchar();
	inter_append(&starkoutj,in->time,get_J(in));

	//if (in->time>5e-11) in->stop_start=TRUE;

//if (in->time>0.0) getchar();
	if (get_dump_status(dump_print_text)==TRUE)
	{
		printf("stark step t=%le V=%le current=%le %le\n",in->time,in->Vapplied,cur,in->last_error);
	}



	device_timestep(in);



	stop_start(in);

	if (in->time>0.0) in->dt*=stark_config.stark_dt_mull;

	step++;

	//printf("%lf\n",in->Vapplied);
	//FILE *out=fopen("./stark.dat","a");
	//fprintf(out,"%le %le %le\n",in->time,nx,px);
	//fclose(out);

	if (in->time>stark_config.stark_stop) break;

	if (in->stop==TRUE)
	{
		break;
	}


}while(1);

dump_dynamic_save(in->outputpath,&store);
dump_dynamic_free(&store);

in->go_time=FALSE;

char outpath[1000];

if (get_dump_status(dump_iodump)==TRUE)
{
	struct istruct calmob;
	struct istruct calr;
	struct istruct calFeff;


	inter_copy(&calr,&starkoutstark_mode,TRUE);
	inter_copy(&calFeff,&starkoutstark_mode,TRUE);


	double F_app=sqrt(starkoutstark_mode.data[0]/stark_config.stark_ea_factor);
	//printf("%le\n",starkoutstark_mode.data[0]);
	//double F_end=sqrt(starkoutstark_mode.data[starkoutstark_mode.len-1]/stark_config.stark_ea_factor);
	//double F_change=fabs(F_app-F_end);
	double stark_Np=4.6e24;//F_change*epsilon0*in->epsilonr[0]/in->ylen/Q;

	FILE *fGtot=fopen("./stark_Gtot.dat","w");
	fprintf(fGtot,"%le\n",in->avg_gen);
	fclose(fGtot);

	//printf("average G=%le\n",in->avg_gen);
	//printf("%le %le %le\n",stark_Np,F_end,F_app);
	//getchar();
	double Feff=0.0;

	for (i=0;i<calr.len;i++)
	{
		Feff=sqrt(starkoutstark_mode.data[i]/stark_config.stark_ea_factor);
		calr.data[i]=epsilon0*in->epsilonr[0]*(F_app-Feff)/(Q*stark_Np);
		calFeff.data[i]=Feff;
	}
	sprintf(outpath,"%s%s",in->outputpath,"stark_r_exp.dat");
	inter_save(&calr,outpath);

	inter_copy(&calmob,&calr,TRUE);
	inter_mod(&calr);
	inter_deriv(&calmob,&calr);
	inter_div(&calmob,&calFeff);


	sprintf(outpath,"%s%s",in->outputpath,"stark_mu_exp.dat");
	inter_save(&calmob,outpath);

	inter_free(&calmob);
	inter_free(&calr);
	inter_free(&calFeff);

	//stark j

	sprintf(outpath,"%s%s",in->outputpath,"starkj.dat");
	inter_save(&starkoutj,outpath);

	//distances

	sprintf(outpath,"%s%s",in->outputpath,"stark_r_n.dat");
	inter_sub_double(&starkoutn,starkoutn.data[0]);
	inter_mod(&starkoutn);
	inter_sub_double(&starkoutp,starkoutp.data[0]);
	inter_mod(&starkoutp);
	if (get_dump_status(dump_iodump)==TRUE) inter_save(&starkoutn,outpath);



	sprintf(outpath,"%s%s",in->outputpath,"stark_r_p.dat");
	if (get_dump_status(dump_iodump)==TRUE) inter_save(&starkoutp,outpath);

	inter_add(&starkoutn,&starkoutp);

	sprintf(outpath,"%s%s",in->outputpath,"stark_r_np.dat");
	inter_save(&starkoutn,outpath);
	//Field
	out=fopena(in->outputpath,"stark_field_stark.dat","w");
	fprintf(out,"%le %le\n",fabs(in->Vapplied/in->ylen),starkoutstark_mode.data[0]);
	fclose(out);

}

sprintf(outpath,"%s%s",in->outputpath,"stark_stark_mode.dat");
inter_save(&starkoutstark_mode,outpath);

//inter_mul(&starkoutstark_mode,-1.0);		//to account of reflection

if (inter_get_max(&starkoutstark_mode)!=0)
{
	inter_norm(&starkoutstark_mode,1.0);
}else
{
	double min=inter_get_min(&starkoutstark_mode);
	if (min!=0)
	{
		inter_div_double(&starkoutstark_mode,fabs(min));
	}
}

buffer_malloc(&buf);
buf.y_mul=1.0;
buf.x_mul=1e12;
strcpy(buf.title,"Reflected light - time");
strcpy(buf.type,"xy");
strcpy(buf.x_label,"Time");
strcpy(buf.y_label,"EA");
strcpy(buf.x_units,"$ps$");
strcpy(buf.y_units,"$arb$");
buf.logscale_x=0;
buf.logscale_y=0;
buffer_add_info(&buf);
buffer_add_xy_data(&buf,starkoutstark_mode.x, starkoutstark_mode.data, starkoutstark_mode.len);
buffer_dump(in->outputpath,"stark_rx_mode_norm.dat",&buf);
buffer_free(&buf);


sprintf(outpath,"%s%s",in->outputpath,"sim_info.dat");
FILE *points=fopen(outpath,"w");
fprintf(points,"#point\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,5e-11));
fprintf(points,"#stark_point0\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,4e-12));
fprintf(points,"#stark_point1\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,40e-12));
fprintf(points,"#stark_point2\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,100e-12));
fprintf(points,"#stark_point3\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,150e-12));
fprintf(points,"#stark_point4\n");
fprintf(points,"%le\n",inter_get_noend(&starkoutstark_mode,180e-12));
fprintf(points,"#light_energy\n");
fprintf(points,"%le\n",6.62606957e-34*3e8/stark_config.stark_probe_wavelength/1.6e-19);
fprintf(points,"#end\n");
fclose(points);

//struct istruct file;
//inter_load_a(&file,"stark_orig.dat",0,1);
//for (i=0;i<file.len;i++)
//{
//starkoutstark.data[i]-=file.data[i];
//}

//sprintf(outpath,"%s%s",in->outputpath,"stark_stark_sub.dat");
//inter_save(&starkoutstark,outpath);

//inter_free(&file);

sprintf(outpath,"%s%s",in->outputpath,"./point_mun.dat");
out=fopen(outpath,"w");
fprintf(out,"%le\n0.0\n",get_avg_mue(in));
fclose(out);

sprintf(outpath,"%s%s",in->outputpath,"./point_mup.dat");
out=fopen(outpath,"w");
fprintf(out,"%le\n0.0\n",get_avg_muh(in));
fclose(out);


inter_free(&starkoutn);
inter_free(&starkoutp);
inter_free(&starkoutj);
inter_free(&starkoutstark_mode);
light_free(&modes);
inter_free(&mymode);
inter_free(&alpha_reduce);
inter_free(&n_reduce);
}

void stark_load_config(struct stark *in,struct device *dev)
{
double ver=0.0;
char name[200];
FILE *config=fopena(dev->outputpath,"stark.inp","r");


config_read_line_to_double(&(in->stark_laser_width),config,"#stark_laser_width");

config_read_line_to_double(&(in->stark_tstart),config,"#stark_start_time");

config_read_line_to_double(&(in->stark_ea_factor),config,"#stark_ea_factor");

config_read_line_to_double(&(in->stark_Np),config,"#stark_Np");

config_read_line_to_double(&(in->stark_den),config,"#stark_den");
in->stark_den=fabs(in->stark_den);

config_read_line_to_double(&(in->stark_externalv),config,"#stark_externalv");

config_read_line_to_double(&(in->stark_dt_neg_time),config,"#stark_dt_neg_time");

config_read_line_to_double(&(in->stark_dt),config,"#stark_dt");

config_read_line_to_double(&(in->stark_dt_mull),config,"#stark_dt_mull");

config_read_line_to_double(&(in->stark_stop),config,"#stark_stop");

config_read_line_to_double(&(in->stark_stark),config,"#stark_stark");
in->stark_stark=fabs(in->stark_stark);

config_read_line_to_double(&(in->stark_lasereff),config,"#stark_lasereff");
in->stark_lasereff=fabs(in->stark_lasereff);

config_read_line_to_double(&(in->stark_probe_wavelength),config,"#stark_probe_wavelength");
in->stark_probe_wavelength=fabs(in->stark_probe_wavelength);

config_read_line_to_int(&(in->stark_sim_contacts),config,"#stark_sim_contacts");
in->stark_sim_contacts=fabs(in->stark_sim_contacts);

config_read_line_to_double(&(in->stark_saturate),config,"#stark_saturate");
in->stark_saturate=fabs(in->stark_saturate);

config_read_line_to_double(&(in->stark_saturate_n),config,"#stark_saturate_n");


config_read_line_to_double(&(ver),config,"#ver");


if (ver!=1.54)
{
ewe("File compatability problem %s\n","stark.inp");
}

unused=fscanf(config,"%s",name);
if (strcmp("#end",name)!=0)
{
ewe("Problem with tof.inp file last item read %s!\n",name);
}
fclose(config);
}
