// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <util.h>
#include <math.h>

#ifndef windows
	#include <sys/socket.h>
	#include <openssl/md5.h>
	#include <sys/types.h>
	#include <sys/inotify.h>
	#include <sys/time.h>
	#include <unistd.h>
	#include <signal.h>
	#include <sys/ioctl.h>
	#include <net/if.h>
	#include <unistd.h>
	#include <netinet/in.h>
	#include <sys/types.h>
	#include <sys/stat.h>
	#define EVENT_SIZE  ( sizeof (struct inotify_event) )
	#define EVENT_BUF_LEN     ( 1024 * ( EVENT_SIZE + 16 ) )
#endif
#include "../../sim.h"
#include "../../server.h"
#include "../../inp.h"
#include "../../timer.h"

static int unused __attribute__((unused));




static int alarm_time=100;
static int stall_time=100;
static int server_jobs_run=0;
static double server_jobs_per_s=0.0;
static int server_odes=0;
static double server_odes_per_s=0.0;

static time_t last_job_ended_at;
static struct server *pserver;

void server_update_last_job_time()
{
last_job_ended_at=time(NULL);
}

void change_cpus(struct server *myserver)
{
#ifndef windows
if (myserver->steel==TRUE)
{
	double load[3];
	getloadavg(load, 3);
	int ncpu=sysconf(_SC_NPROCESSORS_CONF);
	double free_load=ncpu-load[0];
	//printf("%le %le\n",,free_load);
	int add=0;
	if (round(free_load)>0)
	{
	if (ncpu>myserver->cpus) add=1;
	}
	if (round(free_load)<0)
	{
	if (myserver->cpus>myserver->min_cpus) add=-1;
	}

	myserver->cpus+=add;
	printf("CPU: taken=%d delta=%d min=%d max=%d load=%lf\n",myserver->cpus,add,myserver->min_cpus,ncpu,load[0]);
}
#endif
}

void alarm_wakeup (int i)
{
#ifndef windows
	struct itimerval tout_val;

	signal(SIGALRM,alarm_wakeup);

	tout_val.it_interval.tv_sec = 0;
	tout_val.it_interval.tv_usec = 0;
	tout_val.it_value.tv_sec = alarm_time; /* 10 seconds timer */
	tout_val.it_value.tv_usec = 0;

	int since_last_job=(int)(time(NULL)-last_job_ended_at);
	printf("\n>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>Time since %d seconds since last job run.\n",since_last_job);
	if (since_last_job>stall_time)
	{
		ewe("\nJob stalled with %d seconds since last job run.\n",since_last_job);
	}

	setitimer(ITIMER_REAL, &tout_val,0);
	server_check_wall_clock(pserver);
#endif
}


int cmp_lock(char *in)
{
char temp[1000];
int max=strlen("lock");
if (strlen(in)>max)
{
strcpy(temp,in);
temp[max]=0;
return strcmp("lock",temp);
}
return -1;
}

void server_shut_down(struct server *myserver, char *lock_file)
{
FILE *out;
#ifndef windows
if (myserver->cpus!=0)
{
	inotify_rm_watch( myserver->fd, myserver->wd );
	close( myserver->fd );
}
#endif
if (strcmp(lock_file,"")!=0)
{
	out=fopen(lock_file,"w");
	if (out!=NULL)
	{
	fprintf(out,"opvdm lock file\n");
	fclose(out);
	}else
	{
		ewe("Error with lock file %s",lock_file);
	}
}

printf("I have shut down the server\n");
myserver->cpus=0;
myserver->on=FALSE;

}

void print_jobs(struct server *myserver)
{
int i;

for (i=0;i<server_max;i++)
{
	if (myserver->state[i]!=0)
	{
		printf("%d %s %s %d\n",i,myserver->command[i],myserver->output[i],myserver->state[i]);
	}
}
}

double server_get_odes_per_s()
{
return server_odes_per_s;
}

double server_get_jobs_per_s()
{
return server_jobs_per_s;
}

void server_init(struct server *myserver)
{
pserver=myserver;
int i;
int auto_cpus;
char name[200];
double ver;
myserver->on=TRUE;
struct inp_file inp;
#ifndef windows
struct itimerval tout_val;
#endif
int temp=FALSE;
if (myserver->readconfig==TRUE)
{
	#ifndef windows
	inp_init(&inp);
	inp_load_from_path(&inp,"./","server.inp");
	inp_check(&inp,1.21);

	inp_search_int(&inp,&(myserver->cpus),"#server_cpus");

	inp_search_int(&inp,&(stall_time),"#server_stall_time");

	inp_search_int(&inp,&(temp),"#server_exit_on_dos_error");
	inp_search_int(&inp,&(myserver->max_run_time),"#server_max_run_time");

	inp_search_int(&inp,&(auto_cpus),"#server_auto_cpus");
	inp_search_int(&inp,&(myserver->min_cpus),"#server_min_cpus");
	inp_search_int(&inp,&(myserver->steel),"#server_steel");

	inp_free(&inp);
	#endif

	int temp_dump=get_dump_status(dump_lock);
	set_dump_status(dump_lock, FALSE);
	set_dump_status(dump_exit_on_dos_error, temp);
	set_dump_status(dump_lock, temp_dump);

	#ifndef windows
	tout_val.it_interval.tv_sec = 0;
	tout_val.it_interval.tv_usec = 0;
	tout_val.it_value.tv_sec = alarm_time; /* 10 seconds timer */
	tout_val.it_value.tv_usec = 0;
	setitimer(ITIMER_REAL, &tout_val,0);

	signal(SIGALRM,alarm_wakeup); /* set the Alarm signal capture */

	#endif

	textcolor(fg_green);
	printf("Running on %d CPUs\n",myserver->cpus);
	textcolor(fg_reset);
}


myserver->start_time = time(NULL);
myserver->end_time=myserver->start_time+myserver->max_run_time;

struct tm tm = *localtime(&myserver->start_time);
printf("Start time: %d-%d-%d %d:%d:%d\n", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);

tm = *localtime(&myserver->end_time);
printf("End time: %d-%d-%d %d:%d:%d\n", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);

myserver->jobs=0;
myserver->jobs_running=0;

#ifndef windows
if (myserver->cpus>0)
{
	for (i=0;i<server_max;i++)
	{
		myserver->state[i]=server_no_job;
		strcpy(myserver->command[i],"");
		strcpy(myserver->output[i],"");
	}


	myserver->fd=inotify_init();

	if ( myserver->fd < 0 )
	{
		perror( "inotify_init" );
	}

	myserver->wd = inotify_add_watch( myserver->fd, "./", IN_CREATE | IN_CLOSE_WRITE );
}
#endif
}


int server_decode(char *command,char *output)
{
int odes=0;
	if (cmpstr_min(command,"gendosn")==0)
	{
		gen_dos_fd_gaus_n(extract_str_number(command,"gendosn"));
	}else
	if (cmpstr_min(command,"gendosp")==0)
	{
		gen_dos_fd_gaus_p(extract_str_number(command,"gendosp"));
	}else
	{
		odes=run_simulation(command,output);
	}
return odes;
}

void server_add_job(struct server *myserver,char *command,char *output)
{

if (myserver->on==FALSE)
{
	ewe("No server running!\n");
}
#ifndef windows
if (myserver->cpus>0)
{
int i;
	for (i=0;i<server_max;i++)
	{
		if (myserver->state[i]==server_no_job)
		{
			myserver->state[i]=server_job_ready;
			strcpy(myserver->command[i],command);
			strcpy(myserver->output[i],output);
			myserver->jobs++;
			break;
		}
	}
}else
#endif
{
	server_decode(command,output);
	server_update_last_job_time();
}

}

void server_exe_jobs(struct server *myserver)
{
if (myserver->jobs==0) return;
#ifndef windows
if (myserver->jobs_running>=myserver->cpus) return;
int i;

	for (i=0;i<server_max;i++)
	{
		if (myserver->state[i]==server_job_ready)
		{

			if (myserver->jobs_running<myserver->cpus)
			{
				myserver->state[i]=server_job_running;
				printf("just run %s %s\n",myserver->command[i],myserver->output[i]);
				myserver->jobs_running++;
				//printf("I would have run\n");
				//printf("%d %d\n",myserver->jobs_running,myserver->cpus);
				//getchar();

				signal(SIGCHLD, SIG_IGN);
				pid_t pID = fork();
				if (pID == 0)
				{
					int ode=0;
					//printf(">>>>>>>>>>>>run\n");
					char lockname[500];
					char my_temp[500];

					sprintf(my_temp,"%s\n%d\n",myserver->command[i],ode);
					sprintf(lockname,"./lock%d.dat",i);
					set_ewe_lock_file(lockname,my_temp);

					ode=server_decode(myserver->command[i],myserver->output[i]);
					//int i=0;
					//do
					//{
					//i++;
					//}while(i<1e6);

					//printf(">>>>>>>>>>>>fail\n");

					FILE *out=fopen(lockname,"w");
					fprintf(out,"%s\n%d\n",myserver->command[i],ode);
					fclose(out);

					exit(0);
				}
			}else
			{
				break;
			}



		}
	}
#endif
}


void server_job_finished(struct server *myserver,char *job)
{

int i;
for (i=0;i<server_max;i++)
{
	if (strcmp(myserver->command[i],job)==0)
	{
		if (myserver->state[i]!=server_no_job)
		{
			myserver->state[i]=server_no_job;
			strcpy(myserver->command[i],"");
			strcpy(myserver->output[i],"");
			myserver->jobs_running--;
			myserver->jobs--;
		}
		break;
	}
}

}
int server_run_jobs(struct server *myserver)
{
double start_time=timer_get_time();
double stop_time=0.0;
double time_elapsed=0.0;
int length;
int i=0;
int ret=0;
if (myserver->jobs==0) return 0;

#ifndef windows
char buffer[EVENT_BUF_LEN];
if (myserver->cpus>0)
{
	printf("server_run_jobs\n");
	server_exe_jobs(myserver);

do
{
	length = read( myserver->fd, buffer, EVENT_BUF_LEN );

	if ( length < 0 )
	{
		perror( "read" );
	}

	while ( i < length )
	{

		struct inotify_event *event = ( struct inotify_event * ) &buffer[ i ];
		if ( event->len )
		{
			if ( event->mask & IN_CLOSE_WRITE )
			{
					if (cmp_lock(event->name)==0)
					{
						char buf[200];
						//FILE *mes=fopen("./mes.dat","a");
						//fprintf(mes,"\nrecived message: %s\n\n", buf);
						//fclose(mes);
						//getchar();
						FILE *lock=fopen(event->name,"r");
						int ode;
						unused=fscanf(lock,"%s",buf);
						unused=fscanf(lock,"%d",&ode);
						fclose(lock);

						server_odes+=ode;
						server_jobs_run+=1;
						server_job_finished(myserver,buf);
						server_exe_jobs(myserver);

						server_update_last_job_time();

						print_jobs(myserver);
						if (myserver->jobs==0)
						{
							stop_time=timer_get_time();
							time_elapsed=stop_time-start_time;
							server_jobs_per_s=((double)server_jobs_run)/((double)time_elapsed);
							server_odes_per_s=((double)server_odes)/((double)time_elapsed);
							server_odes=0.0;
							server_jobs_run=0.0;
							printf("All jobs finished jobs/s=%lf ode/s=%lf %s\n",server_jobs_per_s,server_odes_per_s,buf);
							return 0;
						}

						if (myserver->jobs_running==0)
						{
							printf("I have finished all my jobs\n");
							break;
						}
					}

					if (strcmp("server_stop.dat",event->name)==0)
					{
						printf("The server has detected a crash\n");
						ret=1;
					}
			}
		}
		i += EVENT_SIZE + event->len;



	}
	i=0;




}while(1);

}
#endif
printf("exit server\n");
return ret;
}

void server_check_wall_clock(struct server *myserver)
{
time_t now = time(NULL);
if (now>myserver->end_time)
{
struct tm tm = *localtime(&now);

printf("Server quit due to wall clock at: %d-%d-%d %d:%d:%d\n", tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
now-=myserver->start_time;
printf("I have run for: %lf hours\n",now/60.0/60.0);
exit(0);
}
}

void server_stop_and_exit()
{
	FILE *out=fopen("./server_stop.dat","w");
	fprintf(out,"solver\n");
	fclose(out);
	exit(0);
}
