// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "../../sim.h"
#include "../../ntricks.h"
#include "../../inp.h"
#include "tof.h"
#include "../../buffer.h"
#include "../../dynamic_store.h"

static int unused __attribute__((unused));

struct tof tof_config;

void sim_tof(struct device *in)
{
struct buffer buf;
buffer_init(&buf);

double V=0.0;
int ittr=0;
int step=0;

tof_load_config(&tof_config,in);
printf("Running ToF. %e\n",tof_config.tof_voltage);

in->go_time=FALSE;

struct dynamic_store store;
dump_dynamic_init(&store,in);

in->dt=tof_config.tof_dt;

light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);

if (fabs(tof_config.tof_voltage-in->Vapplied)>0.2)
{
	ramp_externalv(in,0.0,tof_config.tof_voltage);
}


sim_externalv(in,tof_config.tof_voltage);

tof_config.tof_power_suply=tof_config.tof_voltage+get_equiv_I(in)*tof_config.tof_Rshort;

time_init(in);
device_timestep(in);

in->time=-tof_config.start_steps*in->dt;

printf("%le\n",tof_config.tof_power_suply);

struct istruct tofout;
inter_init(&tofout);

in->go_time=TRUE;
V=in->Vapplied;
double ratio=0.0;
//int dump_count=0;

carrier_count_reset(in);

do
{
	if (step<tof_config.start_steps)
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		reset_np_save(in);
	}else
	if ((step>=tof_config.start_steps)&&(in->time<(tof_config.tof_laser_width)))
	{
		ratio=light_convert_density(in,0.0,tof_config.tof_laser_width);
		light_solve_and_update(in,&(in->mylight),in->Psun,tof_config.toflaserpower*ratio,tof_config.tof_laser_width);
		stop_start(in);
	}else
	{
		light_solve_and_update(in,&(in->mylight),in->Psun,0.0,0.0);
		stop_start(in);
	}

	in->Vapplied=V;

	double i0=newton_sim_tof_all(in,tof_config.use_capacitor);

	if (get_dump_status(dump_print_converge)==TRUE)
	{
	printf("tof time=%e step=%d %.1e ",in->time,step,in->last_error);
	printf("V=%lf current = %e mA (%e A/m^2)\n",V,i0/1e-3,get_J(in));
	}



	plot_now(in,"tof.plot");
	dump_dynamic_add_data(&store,in,in->time);

	dump_write_to_disk(in);

	ittr++;

	if (in->time>0.0) in->dt*=tof_config.tofdtmull;

	inter_append(&tofout, in->time, fabs(i0));

	device_timestep(in);
	step++;

	if (in->time>tof_config.tofstop) break;

	if (in->stop==TRUE)
	{
		break;

	}

	stop_start(in);

}while(1);


in->go_time=FALSE;

char name[200];

inter_sub_double(&tofout,tofout.data[tofout.len-1]);

buffer_malloc(&buf);
buf.y_mul=1.0;
buf.x_mul=1.0;
strcpy(buf.title,"ToF current with DC removed");
strcpy(buf.type,"xy");
strcpy(buf.x_label,"Time");
strcpy(buf.y_label,"Current");
strcpy(buf.x_units,"s");
strcpy(buf.y_units,"A");
buf.logscale_x=1;
buf.logscale_y=1;
buffer_add_info(&buf);
buffer_add_xy_data(&buf,tofout.x, tofout.data, tofout.len);
buffer_dump(in->outputpath,"tof_i.dat",&buf);
buffer_free(&buf);


inter_chop(&tofout,1e-9, 100.0);

int i;

for (i=0;i<tofout.len;i++)
{
	tofout.data[i]*=tofout.x[i];
	tofout.x[i]=((in->Te[0]*kb/Q)*log(tof_config.tof_cap_escape_fx*tofout.x[i]));
}

buffer_malloc(&buf);
buf.y_mul=1.0;
buf.x_mul=1.0;
strcpy(buf.title,"ToF current with DC removed");
strcpy(buf.type,"xy");
strcpy(buf.x_label,"Energy");
strcpy(buf.y_label,"DoS");
strcpy(buf.x_units,"eV");
strcpy(buf.y_units,"au");
buf.logscale_x=0;
buf.logscale_y=0;
buffer_add_info(&buf);
buffer_add_xy_data(&buf,tofout.x, tofout.data, tofout.len);
buffer_dump(in->outputpath,"tof_dos.dat",&buf);
buffer_free(&buf);



dump_dynamic_save(in->outputpath,&store);
dump_dynamic_free(&store);
}

void tof_load_config(struct tof *in,struct device *dev)
{
struct inp_file inp;
inp_init(&inp);
inp_load_from_path(&inp,dev->inputpath,"tof.inp");
inp_check(&inp,1.11);
inp_search_double(&inp,&(in->tofstop),"#tofstop");
inp_search_double(&inp,&(in->tof_laser_width),"#tof_laser_width");
inp_search_double(&inp,&(in->tof_dt),"#dt");
inp_search_double(&inp,&(in->tofdtmull),"#tofdtmull");
inp_search_double(&inp,&(in->toflaserpower),"#toflaserpower");
inp_search_double(&inp,&(in->start_steps),"#start_steps");
inp_search_double(&inp,&(in->tof_voltage),"#tof_voltage");
inp_search_double(&inp,&(in->tof_Rshort),"#tof_Rshort");
inp_search_double(&inp,&(in->tof_cap_escape_fx),"#tof_cap_escape_fx");
inp_search_int(&inp,&(in->use_capacitor),"#use_capacitor");

inp_free(&inp);

}
